""" File handling utilities """

import enum
import itertools
import logging
import os.path

from PySide6.QtCore import QSettings, QStandardPaths, QUrl

from .. import images

LOGGER = logging.getLogger(__name__)
ACCEPT_AUDIO_EXTS = ('.wav', '.ogg', '.flac', '.mp3', '.aif', '.aiff')


class FileRole(enum.Enum):
    """ File roles, for file selector widgets """

    def __init__(self, label, file_filter):
        self.label = label
        self.file_filter = file_filter

    ALBUM = ("album", "Bandcrash album (*.bcalbum *.json)")
    AUDIO = (
        "track", f"Audio files ({' '.join(f'*{ext}' for ext in ACCEPT_AUDIO_EXTS)})")
    IMAGE = (
        "image", f"Image files ({' '.join(f'*{ext}' for ext in images.known_extensions())})")
    STYLESHEET = ("stylesheet", "Stylesheets (*.css)")
    OUTPUT = ("output", '')
    BINARY = ("binary", '')

    @property
    def default_directory(self):
        """ Get the default system-level directory for this file role """
        settings = QSettings()
        settings.beginGroup("defaultDirs")

        if settings.contains(self.name):
            return settings.value(self.name)

        sloc = QStandardPaths.StandardLocation

        for candidate in itertools.chain(
                *[QStandardPaths.standardLocations(loc)
                  for loc in (sloc.MusicLocation, sloc.DocumentsLocation, sloc.HomeLocation)]):
            return candidate

        LOGGER.warning(
            "Couldn't find default directory for role %s", self.name)
        return ''

    @default_directory.setter
    def default_directory(self, file_dir):
        """ Set the default system-level directory for this file role """
        settings = QSettings()
        settings.beginGroup("defaultDirs")
        settings.setValue(self.name, file_dir)
        settings.sync()


def filter_audio_urls(urls: list[QUrl]):
    """ Given a list of audio URLs, provide ones which are a local file in an
    accepted format """

    return [path for path in [url.toLocalFile() for url in urls if url.isLocalFile()]
            if os.path.splitext(path)[1] in ACCEPT_AUDIO_EXTS]



""" Track editing widgets """

import collections
import logging
import os
import os.path
import typing

from PySide6.QtCore import Qt
from PySide6.QtWidgets import (QAbstractItemView, QAbstractScrollArea,
                               QButtonGroup, QCheckBox, QFileDialog,
                               QFormLayout, QHBoxLayout, QLineEdit,
                               QListWidget, QListWidgetItem, QPlainTextEdit,
                               QPushButton, QRadioButton, QScrollArea,
                               QSizePolicy, QSplitter, QVBoxLayout)

from .. import util
from . import datatypes, file_utils
from .file_utils import FileRole
from .widgets import FileSelector, FlowLayout, wrap_layout

LOGGER = logging.getLogger(__name__)


class TrackEditor(QScrollArea):
    """ A track editor pane """
    # pylint:disable=too-many-instance-attributes

    def __init__(self, path_delegate):
        """ edit an individual track

        :param dict data: The metadata blob
        """
        # pylint: disable = too-many-statements
        super().__init__()
        self.setMinimumSize(400, 0)

        self.path_delegate = path_delegate
        self.data: typing.Optional[datatypes.TrackData] = None
        self.setEnabled(False)

        self.setMinimumSize(450, 0)
        self.setWidgetResizable(True)
        self.setSizeAdjustPolicy(
            QAbstractScrollArea.SizeAdjustPolicy.AdjustToContents)

        layout = QFormLayout()
        layout.setFieldGrowthPolicy(
            QFormLayout.FieldGrowthPolicy.AllNonFixedFieldsGrow)
        self.setWidget(wrap_layout(layout))

        self.filename = FileSelector(FileRole.AUDIO, path_delegate)
        self.group = QLineEdit()
        self.title = QLineEdit()
        self.genre = QLineEdit()
        self.artist = QLineEdit()
        self.composer = QLineEdit()
        self.cover_of = QLineEdit()
        self.artwork = FileSelector(FileRole.IMAGE, path_delegate)
        self.lyrics = QPlainTextEdit()
        self.comment = QLineEdit()
        self.about = QPlainTextEdit()
        self.slug = QLineEdit()

        layout.addRow("Audio file", self.filename)
        layout.addRow("Title", self.title)

        self.preview = QRadioButton("Preview")
        self.listed = QRadioButton("Listed")
        self.hidden = QRadioButton("Hidden")

        self.track_type = QButtonGroup()
        self.track_type.addButton(self.preview)
        self.track_type.addButton(self.listed)
        self.track_type.addButton(self.hidden)

        player_options = FlowLayout()
        player_options.setContentsMargins(0, 0, 0, 0)
        player_options.addWidget(self.preview)
        player_options.addWidget(self.listed)
        player_options.addWidget(self.hidden)
        layout.addRow("Track type", player_options)

        self.explicit = QCheckBox("Explicit")
        player_options.addWidget(self.explicit)

        layout.addRow("Genre", self.genre)
        layout.addRow("Grouping", self.group)

        layout.addRow("Track artist", self.artist)
        layout.addRow("Composer", self.composer)
        layout.addRow("Cover of", self.cover_of)
        layout.addRow("Artwork", self.artwork)
        layout.addRow("Lyrics", self.lyrics)
        layout.addRow("Track comment", self.comment)
        layout.addRow("Track details", self.about)
        layout.addRow("Link slug", self.slug)

    def reset(self, data: datatypes.TrackData):
        """ Reset to the specified backing data """
        self.data = data
        self.setEnabled(data is not None)

        for key, widget in (
            ('filename', self.filename.file_path),
            ('title', self.title),
            ('genre', self.genre),
            ('artist', self.artist),
            ('composer', self.composer),
            ('cover_of', self.cover_of),
            ('artwork', self.artwork.file_path),
            ('group', self.group),
            ('comment', self.comment),
        ):
            widget.setText(self.data.get(key, ''))

        self.lyrics.document().setPlainText(
            util.text_to_lines(self.data.get('lyrics', '')))
        self.about.document().setPlainText(util.text_to_lines(self.data.get('about', '')))

        hidden = self.data.get('hidden', False)
        preview = self.data.get('preview', True) and not hidden
        listed = not hidden and not preview
        LOGGER.debug("hidden=%s preview=%s listed=%s", hidden, preview, listed)
        self.hidden.setChecked(hidden)
        self.preview.setChecked(preview)
        self.listed.setChecked(listed)

        self.explicit.setCheckState(
            datatypes.to_checkstate(self.data.get('explicit', False)))

    def apply(self):
        """ Apply our data to the backing data """
        # pylint:disable=too-many-branches

        if not self.data:
            LOGGER.debug("TrackEditor apply - no data")
            return

        LOGGER.debug("TrackEditor.apply %s", self.data.get('filename'))

        relpath = util.make_relative_path(self.path_delegate.filename)

        datatypes.apply_text_fields(self.data, (
            ('filename', self.filename.file_path),
            ('artwork', self.artwork.file_path),
        ), relpath)

        datatypes.apply_text_fields(self.data, (
            ('title', self.title),
            ('genre', self.genre),
            ('artist', self.artist),
            ('composer', self.composer),
            ('cover_of', self.cover_of),
            ('group', self.group),
            ('comment', self.comment),
        ))

        def split_lines(text):
            lines = text.split('\n')
            return lines if len(lines) != 1 else text

        for key, widget in (
            ('lyrics', self.lyrics),
            ('about', self.about),
        ):
            lines = split_lines(widget.document().toPlainText())
            if lines:
                self.data[key] = lines
            elif key in self.data:
                del self.data[key]

        datatypes.apply_radio_fields(self.data, (
            ('preview', self.preview, True),
            ('hidden', self.hidden, False),
        ))

        datatypes.apply_checkbox_fields(self.data, (
            ('explicit', self.explicit, False),
        ))

        LOGGER.debug("applied: %s", self.data)

    def update_placeholders(self, album_data):
        """ Update the placeholder text in the track editor widgets """
        if album_data:
            for field, widget in (
                ('genre', self.genre),
                ('artist', self.artist),
                ('composer', self.composer),
            ):
                LOGGER.debug("updated placeholder for %s", field)
                widget.setPlaceholderText(album_data.get(
                    field, 'If different than album'))


class TrackListEditor(QSplitter):
    """ The track listing panel and editor """
    # pylint:disable=too-many-instance-attributes

    class TrackItem(QListWidgetItem):
        """ an item in the track listing """

        def __init__(self, track_num: int, track: datatypes.TrackData):
            super().__init__()
            self.track_number = track_num
            self.track_data = track
            self.update_name()

        def set_track_num(self, track_num: int):
            """ Update the track number for this one """
            self.track_number = track_num
            self.update_name()

        def reset(self, track_num: int, data: datatypes.TrackData):
            """ Reset the track listing from a new tracklist

            :param list data: album['data']
            """
            LOGGER.debug("TrackItem.reset %s", self.display_name)
            self.track_number = track_num
            self.track_data = data
            self.update_name()

        def apply(self):
            """ Apply the GUI values to the backing store """
            LOGGER.debug("TrackItem.apply %s", self.display_name)
            self.update_name()

        def update_name(self):
            """ Update the display name """
            self.setText(self.display_name)

        @property
        def display_name(self):
            """ Get the display name of this track """
            info = self.track_data
            if info and 'title' in info:
                title = info['title']
            elif info and 'filename' in info:
                title = f"({os.path.basename(info['filename'])})"
            else:
                title = "(unknown)"

            return f"{self.track_number + 1}. {title}"

    class TrackList(QListWidget):
        """ The actual track listing panel """

        def __init__(self, parent, delegate):
            super().__init__(parent)
            self.setDragDropMode(QAbstractItemView.DragDropMode.InternalMove)
            self.setAcceptDrops(True)
            self.delegate = delegate

        def dragEnterEvent(self, event):
            LOGGER.debug("dragEnterEvent %s %s", event, event.proposedAction())
            LOGGER.debug("hasurls: %s", event.mimeData().hasUrls())
            if event.proposedAction() == Qt.DropAction.CopyAction and event.mimeData().hasUrls():
                if files := file_utils.filter_audio_urls(event.mimeData().urls()):
                    LOGGER.debug("accepted files: %s", files)
                    event.acceptProposedAction()
            else:
                super().dragEnterEvent(event)

        def dropEvent(self, event):
            if event.proposedAction() == Qt.DropAction.CopyAction and event.mimeData().hasUrls():
                if files := file_utils.filter_audio_urls(event.mimeData().urls()):
                    LOGGER.debug("adding files: %s", files)
                    self.delegate.add_tracks(files)
                    event.acceptProposedAction()
            elif event.proposedAction() == Qt.DropAction.MoveAction:
                self.delegate.record_undo()
                super().dropEvent(event)

            self.delegate.apply_changes()

    ActionDelegate = collections.namedtuple(
        'ActionDelegate', (
            'add_tracks',
            'apply_changes',
            'record_undo',)
    )

    def __init__(self, parent):
        super().__init__(parent)
        LOGGER.debug("TrackListEditor.__init__ %s", parent)

        self.data: datatypes.TrackList = []
        self.path_delegate = parent.path_delegate
        self.album_editor = parent

        left_panel = QVBoxLayout(self)
        left_panel.setSpacing(0)
        left_panel.setContentsMargins(0, 0, 0, 0)
        self.addWidget(wrap_layout(left_panel))

        self.track_listing = TrackListEditor.TrackList(
            self, TrackListEditor.ActionDelegate(add_tracks=self.add_files,
                                                 apply_changes=self.apply,
                                                 record_undo=parent.record_undo))
        left_panel.addWidget(self.track_listing)

        self.button_add = QPushButton("+")
        self.button_add.clicked.connect(self.add_track_button)
        self.button_delete = QPushButton("-")
        self.button_delete.clicked.connect(self.delete_track)
        self.button_move_up = QPushButton("^")
        self.button_move_up.clicked.connect(self.move_up)
        self.button_move_down = QPushButton("v")
        self.button_move_down.clicked.connect(self.move_down)

        buttons = QHBoxLayout(self)
        buttons.setSpacing(0)
        buttons.setContentsMargins(0, 0, 0, 0)
        buttons.addWidget(self.button_add)
        buttons.addWidget(self.button_delete)
        buttons.addStretch(1000)
        buttons.addWidget(self.button_move_up)
        buttons.addWidget(self.button_move_down)
        left_panel.addWidget(wrap_layout(buttons))

        self.track_editor = TrackEditor(parent.path_delegate)
        self.track_editor.setEnabled(False)
        self.addWidget(self.track_editor)

        self.track_listing.currentRowChanged.connect(self.set_item)

        for widget in (self, self.track_listing):
            policy = widget.sizePolicy()
            policy.setVerticalPolicy(QSizePolicy.Policy.Expanding)
            widget.setSizePolicy(policy)

        self.setSizes([1, 10])

    def reset(self, data: datatypes.TrackList):
        """ Reset to the backing storage """
        LOGGER.debug("TrackListEditor.reset")

        current_row = self.track_listing.currentRow()

        if self.track_listing.count() and self.track_listing.count() != len(data):
            LOGGER.warning("Sync error: Track listing had %d, expected %d",
                           self.track_listing.count(), len(data))

        for idx, track in enumerate(data):
            item = typing.cast(TrackListEditor.TrackItem,
                               self.track_listing.item(idx))
            if item:
                item.reset(idx, track)
            else:
                self.track_listing.addItem(
                    TrackListEditor.TrackItem(idx, track))

        while self.track_listing.count() > len(data):
            self.track_listing.takeItem(self.track_listing.count() - 1)

        self.data = data

        if current_row != self.track_listing.currentRow():
            LOGGER.warning("Sync error: list position changed from %d to %d",
                           self.track_listing.currentRow(), current_row)
            self.track_listing.setCurrentRow(current_row)

    def apply(self):
        """ Save any currently-edited track """
        LOGGER.debug("TrackListEditor.apply")
        self.track_editor.apply()

        self.data.clear()
        for row in range(self.track_listing.count()):
            item = typing.cast(TrackListEditor.TrackItem,
                               self.track_listing.item(row))
            item.set_track_num(row)
            item.apply()
            LOGGER.debug("  -- append %s", item.display_name)
            self.data.append(item.track_data)

    def set_item(self, row):
        """ Signal handler for row change """
        LOGGER.debug("TrackListEditor.set_item")
        self.apply()
        item = typing.cast(TrackListEditor.TrackItem,
                           self.track_listing.item(row))
        if item:
            self.track_editor.reset(item.track_data)
            self.track_editor.setEnabled(True)
        else:
            self.track_editor.reset({})
            self.track_editor.setEnabled(False)

    def add_track_button(self):
        """ Prompt to add some tracks """
        LOGGER.debug("TrackListEditor.add_tracks")
        role = FileRole.AUDIO
        LOGGER.debug("filter: %s", role.file_filter)
        filenames, _ = QFileDialog.getOpenFileNames(
            self,
            "Select audio files",
            dir=self.path_delegate.get_last_directory(role),
            filter=role.file_filter)

        if filenames:
            # update the audio role selection path
            ref_file = filenames[0]
            LOGGER.debug("Audio role: using filename %s", ref_file)
            role.default_directory = os.path.dirname(ref_file)
            self.path_delegate.set_last_directory(
                role, os.path.dirname(ref_file))

        self.add_files(filenames)

    @property
    def current_row(self):
        """ The current selected row """
        return self.track_listing.currentRow()

    @current_row.setter
    def current_row(self, idx):
        """ Change the current row """
        self.track_listing.setCurrentRow(idx)

    def add_files(self, filenames):
        """ Accepts files into the track listing """
        LOGGER.debug("TrackListEditor.add_files")
        self.album_editor.record_undo()
        for filename in filenames:
            _, title = util.guess_track_title(filename)
            track = {'filename': filename, 'title': title}
            self.track_listing.addItem(
                TrackListEditor.TrackItem(len(self.data), track))
            self.data.append(track)
        self.apply()

    def delete_track(self):
        """ Remove a track """
        LOGGER.debug("TrackListEditor.delete_track")
        self.album_editor.record_undo()
        self.track_listing.takeItem(self.track_listing.currentRow())
        self.apply()

    def select_previous(self):
        """ Select the previous track """
        current_row = self.track_listing.currentRow()
        if current_row > 0:
            self.track_listing.setCurrentRow(current_row - 1)

    def select_next(self):
        """ Select the next track """
        current_row = self.track_listing.currentRow()
        if current_row + 1 < self.track_listing.count():
            self.track_listing.setCurrentRow(current_row + 1)

    def move_up(self):
        """ Move the currently-selected track up in the track listing """
        LOGGER.debug("TrackListEditor.move_up")
        self.album_editor.record_undo()
        row = self.track_listing.currentRow()
        if row > 0:
            dest = row - 1
            item = self.track_listing.takeItem(row)
            self.track_listing.insertItem(dest, item)
            self.track_listing.setCurrentRow(dest)

    def move_down(self):
        """ Move the currently-selected track up in the track listing """
        LOGGER.debug("TrackListEditor.move_down")
        self.album_editor.record_undo()
        row = self.track_listing.currentRow()
        if row < self.track_listing.count() - 1:
            dest = row + 1
            item = self.track_listing.takeItem(row)
            self.track_listing.insertItem(dest, item)
            self.track_listing.setCurrentRow(dest)



""" Album encoder dialog """
import collections
import concurrent.futures
import itertools
import logging

from PySide6.QtCore import Qt, QTimer, Signal
from PySide6.QtWidgets import QDialog, QFormLayout, QProgressBar, QPushButton

from .. import process

LOGGER = logging.getLogger(__name__)


class _Encoder(QDialog):
    """ Album encoder dialog box """
    signal = Signal(concurrent.futures.Future)

    def __init__(self, parent, pool, futures):
        super().__init__(parent)
        self.setWindowTitle("Encoding album")
        self.setWindowModality(Qt.WindowModality.WindowModal)

        self.pool = pool
        self.futures = futures

        layout = QFormLayout()

        self.progress: dict[str, QProgressBar] = {}
        for phase, items in futures.items():
            progress_bar = QProgressBar(self)
            progress_bar.setMinimum(0)
            progress_bar.setMaximum(len(items))
            self.progress[phase] = progress_bar
            layout.addRow(phase, progress_bar)

        self.abort = QPushButton("Abort")
        layout.addRow("", self.abort)
        self.abort.clicked.connect(self.stop)

        self.errors = []

        self.signal.connect(self.update_progress)

        self.setLayout(layout)

    def exec_(self):
        LOGGER.debug("overridden exec")
        for future in list(itertools.chain(*self.futures.values())):
            future.add_done_callback(self.signal.emit)

        # If everything finishes before the dialog presents itself, the thing
        # just stalls. So this is a little hack.
        QTimer.singleShot(250, self.check_finished)

        LOGGER.debug("parent exec")
        return super().exec_()

    def stop(self):
        """ End an encode due to error or cancelation """
        LOGGER.warning("Stopping encode")
        self.pool.shutdown(cancel_futures=True)
        self.reject()

    def check_finished(self):
        """ Watchdog to make sure we aren't waiting on an already-complete futures queue """
        for task in list(itertools.chain(*self.futures.values())):
            if not task.done():
                return

        if self.errors:
            self.reject()
        else:
            self.accept()

    def update_progress(self, future):
        """ Update the progress """
        LOGGER.debug("Got update for future %s", future)
        done = True
        for phase, tasks in self.futures.items():
            remaining = len([task for task in tasks if not task.done()])
            LOGGER.debug("%s: %d tasks remaining", phase, remaining)
            if remaining:
                done = False
            self.progress[phase].setValue(len(tasks) - remaining)

        # check to see if the task failed
        if future:
            try:
                future.result()
            except concurrent.futures.CancelledError:
                pass
            except Exception as error:  # pylint:disable=broad-exception-caught
                LOGGER.exception("Got exception %s", error)
                self.errors.append(error)
                self.stop()

        if done:
            LOGGER.info("All tasks finished with %d errors", len(self.errors))
            if self.errors:
                self.reject()
            else:
                self.accept()


def encode(parent, config, album):
    """ Start the album encode and bring up a progress indicator dialog """
    with concurrent.futures.ThreadPoolExecutor(max_workers=config.num_threads) as pool:

        futures: dict[str, list[concurrent.futures.Future]
                      ] = collections.defaultdict(list)

        LOGGER.debug("processing %s", config)
        process(config, album, pool, futures)

        LOGGER.debug("opening dialog")
        dialog = _Encoder(parent, pool, futures)
        LOGGER.debug("waiting for dialog")
        result = dialog.exec_()
        LOGGER.debug("got result %d (%d errors)", result, len(dialog.errors))

    return result, dialog.errors



""" Encoder options """

import dataclasses
import os
import shutil
import typing


def default_path(tool):
    """ Wrapper to keep Sphinx from exposing local build info to the world """
    def get():
        return shutil.which(tool)
    return get


DEFAULT_BUTLER_PATH = shutil.which('butler')


@dataclasses.dataclass
class Options:
    """ Encoder options for processing an album.

    The following parameters are set by whatever is running the album's process.

    :param str input_dir: The working directory for the input specification
    :param str output_dir: The output directory to store the various builds

    :param str butler_path: The path to the itch.io Butler tool. By default,
        searches for ``butler`` in the user's ``PATH``.
    :param str butler_args: Extra arguments to give to the butler tool (e.g. upload
        version, ``--identity``, etc.)

    :param list preview_encoder_args: FFmpeg encoder options for the web player
    :param list mp3_encoder_args: FFmpeg encoder options for the album download
    :param list ogg_encoder_args: FFmpeg encoder options for the Ogg album download
    :param list flac_encoder_args: FFmpeg encoder options for the FLAC album download

    The following parameters are set by the album's specification data, but may be
    overridden by the process runner (e.g. via command line arguments). A value of
    None indicates that it should use the album's configuration; defaults listed
    here are what occurs if it is unset both here and in the album specification.
    If unset, all of them default to True unless otherwise specified.

    :param bool do_preview: Whether to build the web preview
    :param bool do_mp3: Whether to build the MP3 album download
    :param bool do_ogg: Whether to build the Ogg album download
    :param bool do_flac: Whether to build the FLAC album download

    :param bool do_cleanup: Whether to clean up extra files from the target directories
    :param bool do_zip: Whether to build a .zip archive

    :param bool do_butler: Whether to automatically upload builds to
       `itch.io <https://itch.io>`_ via `Butler <https://itch.io/docs/butler/>`_

    :param str butler_target: The Butler target for the upload
        (e.g. ``fluffy/novembeat-2022`` for ``https://fluffy.itch.io/novembeat-2022``)
    :param str butler_prefix: A prefix to add to the Butler channel name; used for
        variations (e.g. ``bob-`` gives channel names of ``bob-mp3``, ``bob-flac``, etc.)
    """
    # pylint:disable=too-many-instance-attributes
    input_dir: typing.Optional[str] = None  # Base directory for all inputs
    output_dir: typing.Optional[str] = None  # Base directory for all outputs

    num_threads: int = os.cpu_count() or 4  # Number of threads to use for encoding

    preview_encoder_args: list[str] = dataclasses.field(
        default_factory="-q:a 5".split().copy)
    mp3_encoder_args: list[str] = dataclasses.field(
        default_factory="-q:a 0".split().copy)
    ogg_encoder_args: list[str] = dataclasses.field(
        default_factory="-q:a 0".split().copy)
    flac_encoder_args: list[str] = dataclasses.field(default_factory=list)

    butler_path: typing.Optional[str] = dataclasses.field(
        default_factory=default_path('butler'))
    butler_args: list[str] = dataclasses.field(default_factory=list)

    # The following options can override the values set in the album specification
    # (thus the None values)

    do_preview: typing.Optional[bool] = None  # Whether to build a web preview
    # Whether to build an MP3 album download
    do_mp3: typing.Optional[bool] = None
    # Whether to build an Ogg album download
    do_ogg: typing.Optional[bool] = None
    # Whether to build a FLAC album download
    do_flac: typing.Optional[bool] = None
    # Whether to build a CD-R bin/cue pair
    do_cdda: typing.Optional[bool] = None

    # Whether to clean up extraneous files from the output directories
    do_cleanup: typing.Optional[bool] = None

    # Whether to build a .zip file of each output
    do_zip: typing.Optional[bool] = None

    # Whether to initiate a Butler upload
    do_butler: typing.Optional[bool] = None

    butler_target: typing.Optional[str] = None  # Butler project target
    # Channel prefix for the Butler uploads
    butler_prefix: typing.Optional[str] = None


def fields():
    """ Get the dataclass fields """
    return dataclasses.fields(Options)



""" Bandcrash GUI """
# pylint:disable=invalid-name,too-few-public-methods,too-many-ancestors
# type: ignore
import argparse
import collections
import concurrent.futures
import copy
import itertools
import json
import logging
import os
import os.path
import shutil
import subprocess
import threading
import typing

from PySide6 import QtCore, QtGui, QtWidgets
from PySide6.QtCore import Qt
from PySide6.QtWidgets import (QApplication, QCheckBox, QDialog, QErrorMessage,
                               QFileDialog, QFormLayout, QFrame, QHBoxLayout,
                               QLabel, QLineEdit, QMainWindow, QMessageBox,
                               QProgressDialog, QPushButton, QSpinBox, QWidget)

from .. import __version__, util
from ..players import camptown
from . import datatypes, encoder, widgets
from .file_utils import FileRole
from .track_editor import TrackListEditor

LOG_LEVELS = [logging.WARNING, logging.INFO, logging.DEBUG]
LOGGER = logging.getLogger(__name__)


def add_menu_item(menu, name, method, shortcut=None, role=None):
    """ Add a menu item """
    action = menu.addAction(name)
    action.triggered.connect(method)
    if shortcut:
        action.setShortcut(QtGui.QKeySequence(shortcut))
    if role:
        action.setMenuRole(role)
    return action


def get_encode_options():
    """ Get the encoder options """

    from .. import options

    LOGGER.debug("get encode options")
    settings = QtCore.QSettings()
    LOGGER.debug("get config defaults")
    config = options.Options()

    for field in options.fields():
        LOGGER.debug("get field %s", field.name)
        if settings.contains(field.name):
            LOGGER.debug("type=%s value=%s", field.type,
                         settings.value(field.name))
            if field.type == list[str]:
                setattr(config, field.name, str(
                    settings.value(field.name)).split())
            else:
                setattr(config, field.name, settings.value(field.name))

    return config


class PreferencesWindow(QDialog):
    """ Sets application-level preferences """
    # pylint:disable=too-many-instance-attributes

    def __init__(self):
        super().__init__()
        LOGGER.debug("Creating prefs window")
        self.setWindowTitle("Bandcrash Preferences")
        self.setMinimumSize(500, 0)

        defaults = get_encode_options()

        def separator():
            frame = QFrame()
            frame.setFrameShape(QFrame.Shape.HLine)
            return frame

        layout = QFormLayout()
        layout.setFieldGrowthPolicy(
            QFormLayout.FieldGrowthPolicy.AllNonFixedFieldsGrow)
        self.setLayout(layout)

        self.num_threads = QSpinBox(self)
        self.num_threads.setMinimum(1)
        self.num_threads.setMaximum(128)
        self.num_threads.setValue(defaults.num_threads)
        layout.addRow("Number of threads", self.num_threads)

        layout.addRow(separator())

        self.preview_encoder_args = QLineEdit()
        self.preview_encoder_args.setText(
            ' '.join(defaults.preview_encoder_args))
        layout.addRow("Preview encoder options", self.preview_encoder_args)

        self.mp3_encoder_args = QLineEdit()
        self.mp3_encoder_args.setText(' '.join(defaults.mp3_encoder_args))
        layout.addRow("MP3 encoder options", self.mp3_encoder_args)

        self.ogg_encoder_args = QLineEdit()
        self.ogg_encoder_args.setText(' '.join(defaults.ogg_encoder_args))
        layout.addRow("Ogg encoder options", self.ogg_encoder_args)

        self.flac_encoder_args = QLineEdit()
        self.flac_encoder_args.setText(' '.join(defaults.flac_encoder_args))
        layout.addRow("FLAC encoder options", self.flac_encoder_args)

        layout.addRow(separator())

        self.butler_path = widgets.FileSelector(
            FileRole.BINARY, text=defaults.butler_path)
        layout.addRow("Butler binary", self.butler_path)
        connect_button = QPushButton("Connect")
        self.butler_path.layout().addWidget(connect_button)
        connect_button.clicked.connect(self.connect_butler)

        buttons = QHBoxLayout()

        reset_button = QPushButton("Load Defaults")
        reset_button.clicked.connect(self.reset_defaults)
        buttons.addWidget(reset_button)

        apply_button = QPushButton("Apply")
        apply_button.clicked.connect(self.accept)
        apply_button.setDefault(True)
        buttons.addWidget(apply_button)

        layout.addRow(buttons)

        self.accepted.connect(self.apply)

    def apply(self):
        """ Save the settings out """
        settings = QtCore.QSettings()
        for key, value in (
            ('num_threads', self.num_threads.value()),

            ('preview_encoder_args', self.preview_encoder_args.text()),
            ('mp3_encoder_args', self.mp3_encoder_args.text()),
            ('ogg_encoder_args', self.ogg_encoder_args.text()),
            ('flac_encoder_args', self.flac_encoder_args.text()),

            ('butler_path', self.butler_path.text()),
        ):
            settings.setValue(key, value)

    def reset_defaults(self):
        """ Reset to defaults """
        from .. import options
        defaults = options.Options()

        LOGGER.debug("foo 1")

        self.num_threads.setValue(defaults.num_threads)
        self.preview_encoder_args.setText(
            ' '.join(defaults.preview_encoder_args))
        self.mp3_encoder_args.setText(' '.join(defaults.mp3_encoder_args))
        self.ogg_encoder_args.setText(' '.join(defaults.ogg_encoder_args))
        self.flac_encoder_args.setText(' '.join(defaults.flac_encoder_args))

        LOGGER.debug("foo 2")

        self.butler_path.setText(defaults.butler_path)

        LOGGER.debug("foo 3")

        self.apply()

    def connect_butler(self):
        """ Connect to butler """
        connection = subprocess.run([self.butler_path.text(), 'login', '--assume-yes'],
                                    capture_output=True,
                                    check=False,
                                    creationflags=getattr(
            subprocess, 'CREATE_NO_WINDOW', 0),
        )
        if connection.returncode:
            QMessageBox.warning(
                self, "Connection failed", connection.stdout.decode())
        else:
            QMessageBox.information(
                self, "Butler connected", connection.stdout.decode())

    @staticmethod
    def show_preferences():
        """ Show a preferences window """
        prefs_window = PreferencesWindow()
        prefs_window.setModal(True)
        prefs_window.exec()

        for window in BandcrashApplication.instance().windows:
            window.apply()


class AlbumEditor(QMainWindow):
    """ An album editor window """
    # pylint:disable=too-many-instance-attributes,too-many-public-methods

    class PathDelegate:
        """ Storage for the album editor's file information """

        def __init__(self, filename):
            self.filename = filename
            self.last_directory = {}

        def get_last_directory(self, role: FileRole, file_path: typing.Optional[str] = None):
            """ Get the last directory used for a file of a particular type

            :param role: The role
            :param str file_path: The current path to use as a reference
            """
            LOGGER.debug("get_last_directory %s %s", role, file_path)
            LOGGER.debug("   %s", self.last_directory)

            if file_path:
                if os.path.isabs(file_path):
                    # We can just use the existing file's directory
                    return os.path.dirname(file_path)

                if self.filename:
                    # Just make it absolute to our directory
                    return os.path.dirname(util.make_absolute_path(self.filename)(file_path))

            if self.filename:
                # We know where we are
                if role.name in self.last_directory:
                    # And we know where the last file of this type was put
                    return util.make_absolute_path(self.filename)(self.last_directory[role.name])
                # just assume the album's directory
                return os.path.dirname(self.filename)

            # We're not mapped to the filesystem, so just use the system default
            return role.default_directory

        def set_last_directory(self, role: FileRole, dir_path: str):
            """ Set the last directory for this role relative to our album file

            :param role: The role
            :param str dir_path: The directory to use as the reference
            """
            LOGGER.debug("set_last_directory %s %s", role, dir_path)
            if self.filename:
                self.last_directory[role.name] = util.make_relative_path(
                    self.filename)(dir_path)
            else:
                # We aren't mapped to the filesystem so let's just stash it as absolute
                self.last_directory[role.name] = dir_path
            LOGGER.debug("   -> %s", self.last_directory[role.name])

    def __init__(self, path: str):
        """ edit an album file

        :param str path: The path to the JSON file
        """
        # pylint:disable=too-many-statements,too-many-locals
        super().__init__()

        self.setMinimumSize(600, 0)

        self.output_dir: typing.Optional[str] = None

        self.undo_history: list[tuple[int, dict[str, typing.Any]]] = []
        self.redo_history: list[tuple[int, dict[str, typing.Any]]] = []

        menubar = self.menuBar()

        file_menu = menubar.addMenu("&File")

        standard_key = QtGui.QKeySequence.StandardKey

        add_menu_item(file_menu, "&New",
                      BandcrashApplication.file_new, standard_key.New)
        add_menu_item(file_menu, "&Open...",
                      BandcrashApplication.file_open, standard_key.Open)
        add_menu_item(file_menu, "&Save", self.save, standard_key.Save)
        add_menu_item(file_menu, "Save &As...",
                      self.save_as, standard_key.SaveAs)
        add_menu_item(file_menu, "&Revert", self.revert, standard_key.Refresh)
        add_menu_item(file_menu, "&Close", self.close, standard_key.Close)

        album_menu = menubar.addMenu("&Album")

        add_menu_item(album_menu, "&Encode", self.encode_album, "Ctrl+Shift+E")

        edit_menu = menubar.addMenu("&Edit")
        self.undo_menu = add_menu_item(
            edit_menu, "&Undo", self.undo_step, standard_key.Undo)
        self.redo_menu = add_menu_item(
            edit_menu, "&Redo", self.redo_step, standard_key.Redo)
        add_menu_item(edit_menu, "&Preferences", PreferencesWindow.show_preferences,
                      standard_key.Preferences,
                      QtGui.QAction.MenuRole.PreferencesRole)

        self.undo_menu.setEnabled(False)
        self.redo_menu.setEnabled(False)

        track_menu = menubar.addMenu("&Track")

        help_menu = menubar.addMenu("&Help")
        add_menu_item(help_menu, "&About...", self.show_about_box, None,
                      QtGui.QAction.MenuRole.AboutRole)
        add_menu_item(help_menu, "&Manual", self.open_manual, None)

        self.path_delegate = AlbumEditor.PathDelegate(path)
        self.save_hash = 0
        self.data: dict[str, typing.Any] = {'tracks': []}
        if path:
            self.reload(path)

        layout = QFormLayout()
        layout.setFieldGrowthPolicy(
            QFormLayout.FieldGrowthPolicy.AllNonFixedFieldsGrow)
        self.setCentralWidget(widgets.wrap_layout(layout))

        self.artist = QLineEdit()
        self.title = QLineEdit()
        self.year = QLineEdit()
        self.year.setValidator(QtGui.QIntValidator(0, 99999))
        self.year.setMaxLength(5)
        self.genre = QLineEdit()
        self.artwork = widgets.FileSelector(FileRole.IMAGE, self.path_delegate)
        self.composer = QLineEdit()

        layout.addRow("Artist", self.artist)
        layout.addRow("Title", self.title)

        hbox = widgets.QHBoxLayout()
        hbox.setContentsMargins(0, 0, 0, 0)
        hbox.setSpacing(0)
        hbox.addWidget(QLabel("Artist"))
        self.artist_url = QLineEdit()
        self.artist_url.setPlaceholderText("https://my-cool-band.com/")
        hbox.addWidget(self.artist_url)
        hbox.addWidget(QLabel("Album"))
        self.album_url = QLineEdit()
        self.album_url.setPlaceholderText(
            "https://my-cool-band.com/my-cool-album")
        hbox.addWidget(self.album_url)
        layout.addRow("URLs", hbox)

        theme_settings = widgets.FlowLayout()
        self.theme_foreground = widgets.ColorPicker(self, "Foreground")
        self.theme_background = widgets.ColorPicker(self, "Background")
        self.theme_highlight = widgets.ColorPicker(self, "Highlight")
        self.hide_footer = QCheckBox("Hide footer")
        theme_settings.addWidget(self.theme_foreground)
        theme_settings.addWidget(self.theme_background)
        theme_settings.addWidget(self.theme_highlight)
        theme_settings.addWidget(self.hide_footer)
        layout.addRow("Theme settings", theme_settings)

        self.user_css = widgets.FileSelector(
            FileRole.STYLESHEET, self.path_delegate)
        layout.addRow("User CSS", self.user_css)

        layout.addRow("Composer", self.composer)
        layout.addRow("Year", self.year)
        layout.addRow("Genre", self.genre)

        layout.addRow("Artwork", self.artwork)

        self.track_listing = TrackListEditor(self)
        layout.addRow("Audio Tracks", QWidget(self))
        layout.addRow(self.track_listing)

        add_menu_item(track_menu, "&Add...",
                      self.track_listing.add_track_button, "Ctrl+Shift+A")
        add_menu_item(track_menu, "&Delete",
                      self.track_listing.delete_track, "Ctrl+Del")
        add_menu_item(track_menu, "&Previous",
                      self.track_listing.select_previous, "PgUp")
        add_menu_item(track_menu, "&Next",
                      self.track_listing.select_next, "PgDown")
        add_menu_item(track_menu, "Move &up",
                      self.track_listing.move_up, "Alt+Up")
        add_menu_item(track_menu, "Move &down",
                      self.track_listing.move_down, "Alt+Down")

        checkboxes = widgets.FlowLayout()
        self.do_preview = QCheckBox("Web player")
        self.do_preview.setToolTip("Build a web-based preview player")
        self.do_mp3 = QCheckBox("MP3")
        self.do_mp3.setToolTip("Generate an album in MP3 format")
        self.do_ogg = QCheckBox("Ogg Vorbis")
        self.do_ogg.setToolTip("Generate an album in Ogg Vorbis format")
        self.do_flac = QCheckBox("FLAC")
        self.do_flac.setToolTip("Generate an album in FLAC lossless format")
        self.do_cdda = QCheckBox("CD")
        self.do_cdda.setToolTip("Generate .bin/.cue files for CD replication")
        self.do_cleanup = QCheckBox("Clean up files")
        self.do_cleanup.setToolTip("Remove extraneous/left over files")
        self.do_zip = QCheckBox("Build .zip files")
        self.do_zip.setToolTip("Make a .zip file for each output")
        checkboxes.addWidget(self.do_preview)
        checkboxes.addWidget(self.do_mp3)
        checkboxes.addWidget(self.do_ogg)
        checkboxes.addWidget(self.do_flac)
        checkboxes.addWidget(self.do_cdda)
        checkboxes.addWidget(self.do_cleanup)
        checkboxes.addWidget(self.do_zip)
        layout.addRow("Build options", checkboxes)

        butler_opts = QHBoxLayout()
        self.do_butler = QCheckBox()
        self.butler_target = QLineEdit()
        self.butler_target.setPlaceholderText("username/my-album-name")
        self.do_butler.setToolTip(
            "Automate uploads to itch.io via the butler tool")
        self.butler_prefix = QLineEdit()
        self.butler_prefix.setPlaceholderText("prefix")
        butler_opts.addWidget(self.do_butler)
        butler_opts.addWidget(self.butler_target, 50)
        butler_opts.addWidget(self.butler_prefix, 10)
        layout.addRow("itch.io butler", butler_opts)

        buttons = QHBoxLayout()

        start_button = QPushButton("Encode")
        start_button.clicked.connect(self.encode_album)

        buttons.addWidget(start_button)

        layout.addRow(buttons)

        self.setWindowTitle(self.filename or 'New Album')

        self.reset()

        for widget in (
            self.artist,
            self.title,
            self.genre,
            self.composer
        ):
            widget.textChanged.connect(self.apply)

        self.apply()
        self.update_hash()

    @property
    def filename(self):
        """ The current filename of the album """
        return self.path_delegate.filename

    @filename.setter
    def filename(self, path):
        self.path_delegate.filename = path

    def update_hash(self):
        """ Update the fingerprint hash """
        old_hash = self.save_hash
        self.save_hash = hash(repr(self.data))
        LOGGER.debug("updating hash from %s to %s", old_hash, self.save_hash)

    def unsaved(self):
        """ Returns whether there are unsaved changes """
        current_hash = hash(repr(self.data))
        LOGGER.debug("save_hash=%d cur_hash=%d", self.save_hash, current_hash)
        return self.save_hash != current_hash

    def reload(self, path):
        """ Load from the backing storage """
        with open(path, 'r', encoding='utf8') as file:
            try:
                self.data = typing.cast(dict[str, typing.Any], json.load(file))
                if 'tracks' not in self.data:
                    raise KeyError('tracks')
                self.update_hash()
            except (json.decoder.JSONDecodeError, KeyError, TypeError):
                err = QErrorMessage(self)
                err.showMessage("Invalid album JSON file")
                self.filename = ''
                self.data = {'tracks': []}

    @property
    def theme_colors(self) -> typing.Iterable[tuple[widgets.ColorPicker, str, str]]:
        """ mapping for theme items to color values """
        return (
            (self.theme_foreground, 'foreground', '#000000'),
            (self.theme_background, 'background', '#ffffff'),
            (self.theme_highlight, 'highlight', '#7f0000'),
        )

    def reset(self):
        """ Reset to the saved values """
        LOGGER.debug("AlbumEditor.reset")

        for key, text_field in (
            ('artist', self.artist),
            ('title', self.title),
            ('artist_url', self.artist_url),
            ('album_url', self.album_url),
            ('genre', self.genre),
            ('artwork', self.artwork.file_path),
            ('composer', self.composer),
            ('butler_target', self.butler_target),
            ('butler_prefix', self.butler_prefix),
        ):
            text_field.setText(self.data.get(key, ''))

        if 'year' in self.data:
            self.year.setText(str(self.data['year']))

        self.track_listing.reset(self.data['tracks'])

        self.do_preview.setCheckState(
            datatypes.to_checkstate(self.data.get('do_preview', True)))
        self.do_mp3.setCheckState(
            datatypes.to_checkstate(self.data.get('do_mp3', True)))
        self.do_ogg.setCheckState(
            datatypes.to_checkstate(self.data.get('do_ogg', True)))
        self.do_flac.setCheckState(
            datatypes.to_checkstate(self.data.get('do_flac', True)))
        self.do_cdda.setCheckState(
            datatypes.to_checkstate(self.data.get('do_cdda', False)))
        self.do_zip.setCheckState(
            datatypes.to_checkstate(self.data.get('do_zip', True)))
        self.do_cleanup.setCheckState(
            datatypes.to_checkstate(self.data.get('do_cleanup', True)))
        self.do_butler.setCheckState(
            datatypes.to_checkstate(self.data.get('do_butler', True)))

        theme = self.data.get('theme', self.data.get('blamscamp', {}))
        for color, key, dfl in self.theme_colors:
            color.setName(theme.get(key, dfl))
        self.user_css.setText(theme.get('user_css', ''))
        self.hide_footer.setCheckState(
            datatypes.to_checkstate(theme.get('hide_footer', False)))

        self.path_delegate.last_directory = self.data.get(
            '_gui', {}).get('lastdir', {})

    def apply(self):
        """ Apply edits to the saved data """
        LOGGER.debug("AlbumEditor.apply")

        self.record_undo()

        relpath = util.make_relative_path(self.filename)

        datatypes.apply_text_fields(self.data, (
            ('title', self.title),
            ('genre', self.genre),
            ('album_url', self.album_url),
            ('artist_url', self.artist_url),
            ('artist', self.artist),
            ('composer', self.composer),
            ('butler_target', self.butler_target),
            ('butler_prefix', self.butler_prefix),
        ))

        datatypes.apply_text_fields(self.data, (
                                    ('artwork', self.artwork.file_path),
                                    ),
                                    relpath)

        datatypes.apply_text_fields(self.data, (
            ('year', self.year),
        ),
            int)

        datatypes.apply_checkbox_fields(self.data, (
            ('do_preview', self.do_preview, True),
            ('do_mp3', self.do_mp3, True),
            ('do_ogg', self.do_ogg, True),
            ('do_flac', self.do_flac, True),
            ('do_cdda', self.do_cdda, False),
            ('do_zip', self.do_zip, True),
            ('do_cleanup', self.do_cleanup, True),
            ('do_butler', self.do_butler, True),
        ))
        self.track_listing.apply()

        theme = self.data.setdefault('theme', {})
        datatypes.apply_text_fields(theme, (
            ('user_css', self.user_css.file_path),
        ), relpath)
        datatypes.apply_checkbox_fields(theme, (
            ('hide_footer', self.hide_footer, False),
        ))
        for widget, key, dfl in self.theme_colors:
            if widget.name() != dfl:
                theme[key] = widget.name()
            elif key in theme:
                del theme[key]

        datatypes.apply_checkbox_fields(theme, (
            ('hide_footer', self.hide_footer, False),
        ))

        self.data['_gui'] = {
            'lastdir': self.path_delegate.last_directory
        }

        # update whether the itch.io checkbox is enabled
        butler_path = get_encode_options().butler_path
        if bool(butler_path and shutil.which(butler_path)):
            self.do_butler.setEnabled(True)
            self.butler_target.setPlaceholderText("username/my-album-name")
        else:
            self.do_butler.setEnabled(False)
            self.butler_target.setPlaceholderText(
                "Configure butler in the application preferences")

        self.track_listing.track_editor.update_placeholders(self.data)

    @property
    def history_state(self):
        """ Get the current edit history state """
        return self.track_listing.current_row, copy.deepcopy(self.data)

    @history_state.setter
    def history_state(self, state):
        """ Apply an edit history state """
        row, data = state
        self.data = data
        self.reset()
        self.track_listing.current_row = row

    def undo_step(self):
        """ Undo one action """
        if self.undo_history:
            LOGGER.debug("Undoing a step")
            self.redo_history.append(self.history_state)
            self.redo_menu.setEnabled(True)

            self.history_state = self.undo_history.pop()

            self.undo_menu.setEnabled(bool(self.undo_history))
            LOGGER.debug("history size = %d/%d",
                         len(self.undo_history), len(self.redo_history))

    def redo_step(self):
        """ Redo an undone action """
        if self.redo_history:
            LOGGER.debug("Redoing a step")
            self.undo_history.append(self.history_state)
            self.undo_menu.setEnabled(True)

            self.history_state = self.redo_history.pop()

            self.redo_menu.setEnabled(bool(self.redo_history))
            LOGGER.debug("history size = %d/%d",
                         len(self.undo_history), len(self.redo_history))

    def record_undo(self):
        """ Record an undo step """
        LOGGER.debug("Recording undo step")
        self.undo_history.append(self.history_state)
        self.undo_menu.setEnabled(True)
        self.redo_history.clear()
        self.redo_menu.setEnabled(False)
        LOGGER.debug("history size = %d/%d",
                     len(self.undo_history), len(self.redo_history))

    def save(self):
        """ Save the file to disk """
        LOGGER.debug("AlbumEditor.save")
        self.apply()
        if not self.filename:
            return self.save_as()

        with open(self.filename, 'w', encoding='utf8') as file:
            json.dump(self.data, file, indent=3)
            self.update_hash()
        return True

    def save_as(self):
        """ Save the file and change the name """
        LOGGER.debug("AlbumEditor.save_as")
        self.apply()

        role = FileRole.ALBUM
        path, _ = QFileDialog.getSaveFileName(
            self,
            "Select your album file",
            os.path.dirname(self.filename) or role.default_directory,
            role.file_filter,
        )
        if path:
            self.renormalize_paths(self.filename, path)
            self.path_delegate.filename = path
            self.setWindowTitle(self.filename)
            self.reset()
            self.save()
            role.default_directory = os.path.dirname(self.filename)
            return True
        return False

    def revert(self):
        """ Revert all changes """
        LOGGER.debug("AlbumEditor.revert")
        self.apply()

        do_revert = True
        if self.unsaved():
            do_revert = False
            answer = QMessageBox.question(self, "Confirmation",
                                          "Really revert all changes?",
                                          (QMessageBox.StandardButton.Yes |
                                           QMessageBox.StandardButton.No),
                                          QMessageBox.StandardButton.No)

            if answer == QMessageBox.StandardButton.Yes:
                do_revert = True

        if do_revert:
            self.reload(self.filename)
            self.reset()

    def encode_album(self):
        """ Run the encoder process """
        # pylint:disable=too-many-branches,too-many-statements,too-many-locals
        LOGGER.debug("AlbumEditor.encode_album")
        self.apply()

        config = get_encode_options()
        config.input_dir = os.path.dirname(self.filename)

        # find a good default directory to stash the output in
        role = FileRole.OUTPUT
        if not self.output_dir:
            self.output_dir = role.default_directory

        LOGGER.debug("self.output_dir = %s", self.output_dir)

        # prompt for the actual output directory
        dialog = QFileDialog(
            self, "Choose an output directory", self.output_dir or '')
        dialog.setFileMode(QFileDialog.FileMode.Directory)
        dialog.setAcceptMode(QFileDialog.AcceptMode.AcceptOpen)
        dialog.setLabelText(QFileDialog.DialogLabel.Accept, "Encode")
        if not dialog.exec():
            return

        result = dialog.selectedFiles()
        if not result or not result[0]:
            return

        base_dir = result[0]

        # store our output directory for later
        self.output_dir = base_dir or ''

        role.default_directory = self.output_dir

        # Users will most likely be choosing a generic directory and NOT one that's
        # already sandboxed, so, let's make that sandbox for them (just for better UX).
        filename_parts = [part for part in [
            self.data.get(field) for field in ('artist', 'title')]
            if part]
        config.output_dir = os.path.join(
            self.output_dir, util.slugify_filename(' - '.join(filename_parts)))

        config.butler_args = BandcrashApplication.instance().config.butler_args.split()

        LOGGER.info("Album data: %s", self.data)
        LOGGER.info("Config options: %s", config)

        try:
            result, errors = encoder.encode(self, config, self.data)
        except RuntimeError as error:
            QMessageBox.warning(self, "An error occurred", str(error))

        LOGGER.debug("Finished: %d %s", result, errors)

        if errors:
            dlg = widgets.ErrorMessage(self, errors)
            dlg.exec_()
        elif result:
            task_names = "Encode"
            if config.do_butler:
                task_names += " and upload"
            result = QMessageBox.information(
                self,
                "Encode complete",
                f"{task_names} completed successfully",
                QMessageBox.StandardButton.Open | QMessageBox.StandardButton.Ok,
                QMessageBox.StandardButton.Open)  # type:ignore
            if result == QMessageBox.StandardButton.Open:
                QtGui.QDesktopServices.openUrl(
                    QtCore.QUrl.fromLocalFile(config.output_dir))

    def renormalize_paths(self, old_name, new_name):
        """ Renormalize the file paths in the backing data """
        abspath = util.make_absolute_path(old_name)
        relpath = util.make_relative_path(new_name)

        LOGGER.debug("renormalize_paths %s %s", old_name, new_name)

        def renorm(path):
            if os.path.isabs(path):
                LOGGER.debug("Keeping %s absolute", path)
                return path

            old_abs = abspath(path)
            if not os.path.exists(old_abs):
                LOGGER.warning(
                    "Not touching nonexisting path %s (%s)", path, old_abs)
                return path

            out = relpath(old_abs)
            LOGGER.debug("Renormalizing %s -> %s -> %s", path, old_abs, out)
            return out

        for key in ('artwork',):
            if key in self.data:
                self.data[key] = renorm(self.data[key])

        for track in self.data['tracks']:
            for key in ('filename', 'artwork', 'lyrics'):
                if key in track and isinstance(track[key], str):
                    track[key] = renorm(track[key])

        # last_directory is aliased into data instead of being copied
        LOGGER.debug("last_directory before %s",
                     self.path_delegate.last_directory)
        self.path_delegate.last_directory.update({
            key: renorm(value)
            for key, value
            in self.path_delegate.last_directory.items()
        })
        LOGGER.debug("after %s", self.path_delegate.last_directory)

    def show_about_box(self):
        """ Show the about box for the app """
        QMessageBox.about(self, "Bandcrash",
                          f"Bandcrash version {__version__}")

    def closeEvent(self, event):
        self.apply()
        do_close = True
        if self.unsaved():
            do_close = False
            answer = QMessageBox.question(self, "Unsaved changes",
                                          "You have unsaved changes. Really close?",
                                          (QMessageBox.StandardButton.Save |
                                           QMessageBox.StandardButton.Close |
                                           QMessageBox.StandardButton.Cancel))

            if answer == QMessageBox.StandardButton.Save:
                do_close = self.save()
            elif answer == QMessageBox.StandardButton.Close:
                do_close = True

        if do_close:
            BandcrashApplication.instance().release_editor(self)
        else:
            event.ignore()

    @staticmethod
    def open_manual():
        """ Opens the online documentation """
        QtGui.QDesktopServices.openUrl(QtCore.QUrl(
            'https://bandcrash.readthedocs.io'))


class BandcrashApplication(QApplication):
    """ Application event handler """

    def __init__(self, config):
        super().__init__()

        self.config = config

        self.windows: set[AlbumEditor] = set()

        for path in config.open_files:
            self.open_file(os.path.abspath(path))

        QtCore.QTimer.singleShot(100, self.open_on_startup)

    @staticmethod
    def instance():
        """ Get the current app instance """
        return typing.cast(BandcrashApplication, QApplication.instance())

    @staticmethod
    def file_new():
        """ Create a new album file """
        BandcrashApplication.instance().open_file('')

    @staticmethod
    def file_open(or_new: bool = False):
        """ Dialog box to open an existing file

        :param bool or_new: Fallback to a new document if the user cancels
        """
        role = FileRole.ALBUM
        path, _ = QFileDialog.getOpenFileName(None,
                                              "Open album",
                                              role.default_directory,
                                              role.file_filter)
        if path or or_new:
            if path:
                role.default_directory = os.path.dirname(path)
            BandcrashApplication.instance().open_file(path)

    def open_file(self, path):
        """ Open a file into a new window """
        editor = AlbumEditor(path)
        editor.show()
        self.windows.add(editor)

    def release_editor(self, editor):
        """ Release a previously-opened editor """
        self.windows.remove(editor)

    def open_on_startup(self):
        """ Hacky way to open the file dialog on startup. there must be a better way... """
        if not self.windows:
            self.file_open(or_new=True)

    def event(self, evt):
        """ Handle an application-level event """
        LOGGER.debug("Event: %s", evt)
        if evt.type() == QtCore.QEvent.Type.FileOpen:
            LOGGER.debug("Got file open event: %s", evt.file())
            self.open_file(evt.file())
            return True

        return super().event(evt)


def main():
    """ instantiate an app """
    QtCore.QCoreApplication.setOrganizationName("busybee")
    QtCore.QCoreApplication.setOrganizationDomain("beesbuzz.biz")
    QtCore.QCoreApplication.setApplicationName("Bandcrash")

    parser = argparse.ArgumentParser(description="GUI for Bandcrash")
    parser.add_argument('-v', '--verbosity', action="count",
                        help="increase output verbosity", default=0)
    parser.add_argument('--version', action='version',
                        version=f"%(prog)s {__version__}")
    parser.add_argument('open_files', type=str, nargs='*')
    parser.add_argument('--butler-args', type=str, default='',
                        help='Options to pass along to butler push')
    options = parser.parse_args()

    logging.basicConfig(level=LOG_LEVELS[min(
        options.verbosity, len(LOG_LEVELS) - 1)],
        format='%(message)s')

    LOGGER.debug(
        "Opening bandcrash GUI with provided files: %s", options.open_files)

    app = BandcrashApplication(options)

    app.exec()


if __name__ == '__main__':
    main()



""" bin/cue generator """

import concurrent.futures
import logging
import os
import os.path
import re
import struct

from . import util

try:
    from .__version__ import __version__
except ImportError:
    __version__ = '(unknown)'

LOGGER = logging.getLogger(__name__)

BYTES_PER_SAMPLE = 4
SAMPLES_PER_SECOND = 44100
FRAMES_PER_SECOND = 75


def cue_quote(text, force_quote=False):
    """ Quote a string for a cue file """

    # normalize whitespace
    text = ' '.join(str(text).split())

    # replace " with '' because some tools are really silly
    text = text.replace('"', "''")

    # remove nonprintable characters
    text = re.sub(r'[\x00-\x1F\x7F-\x9F]', '', text)

    # quote values with spaces
    if force_quote or ' ' in text:
        text = f'"{text}"'

    return text


def parse_time(samples):
    """ Convert a sample count to minutes/seconds/frames """
    frames = int(samples*FRAMES_PER_SECOND/SAMPLES_PER_SECOND)
    seconds, frames = divmod(frames, FRAMES_PER_SECOND)
    minutes, seconds = divmod(seconds, 60)
    return minutes, seconds, frames


def format_time(samples):
    """ format a timecode in MM:SS:FF """
    mm, ss, ff = parse_time(samples)
    return f'{mm:02}:{ss:02}:{ff:02}'


class CDWriter:
    """ class for building the .bin and .cue file """
    # pylint:disable=too-many-instance-attributes

    def __init__(self, input_dir, output_dir, album, protections, fname='album.bin'):
        # pylint:disable=too-many-positional-arguments,too-many-arguments
        self.input_dir = input_dir
        self.output_dir = output_dir
        self.album = album
        self.protections = protections

        # array of (track_dict, offset, duration)
        self.tracks = []

        self.timecode = 0   # current timecode, in samples

        self.protections.add(fname)
        self.bin_fname = fname
        self.bin_path = os.path.join(self.output_dir, fname)

        # create the empty binfile
        with open(self.bin_path, 'wb') as _:
            pass

    def add_track(self, track):
        """ Add a track to the bin and cue """

        idx = len(self.tracks) + 1
        LOGGER.info("Encoding track %d - %s", idx,
                    track.get('title', '(no title)'))

        start_timecode = self.timecode
        track_duration = 0

        if track.get('filename'):
            infile = os.path.join(self.input_dir, track['filename'])
            target = os.path.join(self.output_dir, f'track-{idx}.tmp.pcm')
            LOGGER.debug("Converting %s to %s as raw", infile, target)
            try:
                util.run_encoder(infile, target, [
                                 '-f', 's16le', '-acodec', 'pcm_s16le', '-ar', '44100'])
                with open(self.bin_path, 'ab') as binfile:
                    with open(target, 'rb') as tempdata:
                        size = 0
                        while chunk := tempdata.read(16384):
                            binfile.write(chunk)
                            size += len(chunk)
                        if size % BYTES_PER_SAMPLE != 0:
                            raise RuntimeError(
                                f"{size} bytes is not a multiple of {BYTES_PER_SAMPLE}")
                        self.timecode += size // BYTES_PER_SAMPLE
                        track_duration = int(
                            size/(BYTES_PER_SAMPLE*SAMPLES_PER_SECOND) + 0.5)
            finally:
                os.remove(target)

        self.tracks.append((track, start_timecode, track_duration))

    def commit(self, leadout=2*SAMPLES_PER_SECOND):
        """ commit the .bin, with a specified leadout size """
        with open(self.bin_path, 'ab') as binfile:
            binfile.write(bytearray(leadout * BYTES_PER_SAMPLE))

    def write_cue(self, fname='album.cue'):
        """ write the text cuefile """
        LOGGER.info("Writing text-format cue file to %s", fname)

        with open(os.path.join(self.output_dir, fname), 'w',
                  encoding='iso-8859-1', errors='replace', newline='\r\n') as cue:
            def writeln(line):
                print(line, file=cue)

            def write_props(props, indent=''):
                """ write an array of properties, formatted as key, value, force_quote """
                for key, value, force_quote in props:
                    if value is not None:
                        writeln(
                            f'{indent}{key} {cue_quote(value, force_quote)}')

            write_props([
                ('PERFORMER', self.album.get('artist'), True),
                ('SONGWRITER', self.album.get('composer'), True),
                ('TITLE', self.album.get('title'), True),
                ('CATALOG', self.album.get('upc'), False),
                ('REM GENRE', self.album.get('genre'), True),
                ('REM DATE', self.album.get('year'), False),
                ('REM COMMENT', f'Authored by Bandcrash {__version__}', True),
            ])

            writeln(f'FILE "{self.bin_fname}" BINARY')

            def get_prop(track, key, fallback=False):
                if not track:
                    return None
                return track.get(key, self.album.get(key) if fallback else None)

            for idx, (track, offset, _) in enumerate(self.tracks, start=1):
                writeln(f"  TRACK {idx:02} AUDIO")
                write_props([
                    ('TITLE', get_prop(track, 'title'), True),
                    ('PERFORMER', get_prop(track, 'artist', True), True),
                    ('SONGWRITER', get_prop(track, 'composer', True), True),
                    ('REM GENRE', get_prop(track, 'genre'), True),
                    ('ISRC', get_prop(track, 'isrc'), False),
                    ('INDEX 01', format_time(offset), False)
                ], indent='    ')

            writeln(f"REM LEAD-OUT {format_time(self.timecode)}")

        self.protections.add(fname)

    def write_kunaki_cue(self, fname='kunaki.cue'):
        """ write the kunaki-format cuefile """
        LOGGER.info("Writing Kunaki-format cue file to %s", fname)

        # Kunaki's proprietary CUE format is semi-documented at
        # https://gist.github.com/LoneRabbit/36f2a74c27a7d6a3b443b44d27fd2702
        # from which this code is adapted. It appears to be a subset of the
        # raw data stored in a CD's TOC.
        #
        # Each track record looks like:
        #
        # <ctrl><adr> <tt> <ii> <xx> 00 <mm> <ss> <ff>
        #
        # where ctrl is 4 bits: quad 0 copy-allowed pre-emphasis
        # adr = 1 for track info, 3 for EOF marker
        # xx = lead in/out? generated pregap maybe?
        # mm/ss/ff: timecode in minutes/seconds/frames (75 frames per second)
        #
        # In Kunaki's generated cuefiles, ctrl|adr is always 01 for tracks, 03 for EOF
        # and xx is 1 for lead-in and lead-out, 0 otherwise
        #
        # The puzzling thing is that their TOC shows the actual track starts
        # as offset by the pregap size, but there is no corresponding pregap
        # within the .bin file (which Kunaki erroneously calls an "iso").
        # Baking in a pregap as one would expect causes the tracks to be offset
        # by that amount. So there seems to be some magic pregap logic happening
        # on Kunaki's side.

        pregap = 2*SAMPLES_PER_SECOND

        with open(os.path.join(self.output_dir, fname), 'wb') as cue:
            def write_item(adr, track, index, xx, timecode):
                mm, ss, ff = parse_time(timecode)
                cue.write(struct.pack('B'*8,
                                      adr, track, index, xx, 0, mm, ss, ff))

            # pregap
            write_item(1, 0, 0, 1, 0)
            write_item(1, 1, 0, 0, 0)

            # each track, offset by the pregap size
            for tnum, (_, offset, _) in enumerate(self.tracks, start=1):
                write_item(1, tnum, 1, 0, offset + pregap)

            # lead-out
            write_item(1, 0xAA, 1, 1, self.timecode + pregap)

            # EOF marker
            cue.write(struct.pack('B', 3))

        self.protections.add(fname)

    def write_tsv(self, fname='tracklist.tsv'):
        """ Write out an easily-parsed track listing file """
        LOGGER.info("Writing TSV file to %s", fname)

        with open(os.path.join(self.output_dir, fname), 'w', encoding='utf-8') as tsv:
            print('Index\tTitle\tDuration\tStart Time\tPerformer\tComposer', file=tsv)

            def get_prop(track, key):
                return ' '.join(track.get(key, self.album.get(key, '')).split())
            for idx, (track, offset, duration) in enumerate(self.tracks, start=1):
                minutes, seconds = divmod(duration, 60)
                print('\t'.join([
                    str(idx),
                    get_prop(track, 'title'),
                    f'{int(minutes)}:{int(seconds+0.5):02}',
                    format_time(offset),
                    get_prop(track, 'artist'),
                    get_prop(track, 'composer')
                ]), file=tsv)

        self.protections.add(fname)


def encode(album, input_dir, output_dir, protections, pool):
    """ Run the bincue output process """
    processor = CDWriter(input_dir, output_dir, album, protections)
    futures: list[concurrent.futures.Future] = []

    def task(wait_for, func, *args):
        if wait_for:
            wait_for.result()
        func(*args)

    def submit(wait_for, func, *args):
        future = pool.submit(task, wait_for, func, *args)
        futures.append(future)
        return future

    last_step = None

    for track in album['tracks']:
        last_step = submit(last_step, processor.add_track, track)

    last_step = submit(last_step, processor.commit)

    # these can all run in parallel, so they don't need to update last_step
    for step in (processor.write_cue,
                 processor.write_kunaki_cue,
                 processor.write_tsv):
        submit(last_step, step)

    return futures



""" main CLI entry point """

import argparse
import collections
import concurrent.futures
import dataclasses
import itertools
import json
import logging
import os
import sys
import typing

from . import __version__, options, process, util

LOG_LEVELS = [logging.WARNING, logging.INFO, logging.DEBUG]
LOGGER = logging.getLogger(__name__)


def parse_args():
    """ Parse commandline arguments """
    defaults = options.Options()
    parser = argparse.ArgumentParser("Bandcrash CLI")

    parser.add_argument('input', type=str,
                        help="Album input directory or JSON file")
    parser.add_argument('output', type=str, nargs='?', default=None,
                        help="Album output directory")

    parser.add_argument("-v", "--verbosity", action="count",
                        help="Increase output logging level", default=0)
    parser.add_argument("--version", action="version",
                        version=f"%(prog)s {__version__}")

    parser.add_argument('--num-threads', '-t', type=int,
                        dest='num_threads',
                        help="Maximum number of concurrent threads",
                        default=defaults.num_threads)

    parser.add_argument('--json', '-j', type=str, dest='json_file',
                        default='album.json',
                        help="If input is a directory, specifies the JSON file "
                        "(relative to that directory)")

    parser.add_argument('--init', dest='init_json', action='store_true',
                        help="Populate the JSON file automatically")

    for target, description, add_args in (
        ('preview', 'Build web player', True),
        ('mp3', 'Encode mp3 album', True),
        ('ogg', 'Encode ogg album', True),
        ('flac', 'Encode flac album', True),
        ('cdda', 'Generate .bin/.cue file for CD-R replication', False),
        ('cleanup', 'Clean extraneous files in the output directories', False),
        ('zip', 'Build a .zip archive', False),
        ('butler', 'Upload to itch.io using Butler', False),
    ):
        feature = parser.add_mutually_exclusive_group(required=False)
        fname = f'do_{target}'
        feature.add_argument(f'--{target}', dest=fname, action='store_true',
                             help=description)
        feature.add_argument(f'--no-{target}', dest=fname, action='store_false',
                             help=f"Don't {description}")
        feature.set_defaults(**{fname: None})

        if add_args:
            parser.add_argument(f'--{target}-encoder-args', type=str,
                                help=f"Arguments to pass to the {target} encoder",
                                default=' '.join(getattr(defaults, f'{target}_encoder_args')))

    parser.add_argument('--butler-path', type=str, default=defaults.butler_path,
                        help="Path to the butler executable")

    parser.add_argument('--butler-target', '-b', type=str,
                        dest='butler_target',
                        help="Butler push target prefix",
                        default=None)

    parser.add_argument('--butler-channel-prefix', '-p', type=str,
                        dest='butler_prefix',
                        help="Prefix for the Butler channel name",
                        default=None)

    parser.add_argument('--butler-args', type=str, default='',
                        help="Extra arguments to provide to butler")

    return parser.parse_args()


def get_config(args) -> options.Options:
    """ Convert the parsed command arguments to an options structure """

    config = options.Options()

    for field in dataclasses.fields(config):
        value = getattr(args, field.name, None)
        if value is not None:
            if field.type == list[str]:
                LOGGER.debug("Setting config list %s to %s", field.name, value)
                setattr(config, field.name, value.split())
            else:
                LOGGER.debug("Setting config field %s to %s",
                             field.name, value)
                setattr(config, field.name, value)

    return config


def main():
    """ Main entry point """
    # pylint:disable=too-many-branches,too-many-statements,too-many-locals
    args = parse_args()
    config = get_config(args)

    logging.basicConfig(level=LOG_LEVELS[min(
        args.verbosity, len(LOG_LEVELS) - 1)],
        format='%(message)s')

    if os.path.isdir(args.input):
        config.input_dir = args.input
        json_file = os.path.join(args.input, args.json_file)
    else:
        config.input_dir = os.path.dirname(args.input)
        json_file = args.input
    assert config.input_dir is not None

    config.output_dir = args.output

    if not os.path.isfile(json_file) and not args.init_json:
        LOGGER.error("%s not found and --init not specified", json_file)
        sys.exit(1)

    if not config.output_dir and not args.init_json:
        LOGGER.error("Output directory not specified")
        sys.exit(1)

    if os.path.isfile(json_file):
        with open(json_file, 'r', encoding='utf8') as file:
            album = json.load(file)
    else:
        album = None

    if args.init_json:
        LOGGER.info("Populating %s with files from %s",
                    json_file, config.input_dir)
        album = util.populate_album(config.input_dir, album)
        LOGGER.info("Album now has %d tracks", len(album['tracks']))
        with open(json_file, 'w', encoding='utf8') as file:
            json.dump(album, file)

    if not config.output_dir:
        return

    pool = concurrent.futures.ThreadPoolExecutor(
        max_workers=config.num_threads)

    futures: typing.Dict[str,
                         typing.List[concurrent.futures.Future]] = collections.defaultdict(list)

    process(config, album, pool, futures)

    all_tasks = list(itertools.chain(*futures.values()))
    remaining_tasks = [f for f in all_tasks if not f.done()]
    LOGGER.info("Waiting for all tasks to complete... (%d/%d pending)",
                len(remaining_tasks), len(all_tasks))

    errors = []
    for task in concurrent.futures.as_completed(all_tasks):
        try:
            task.result()
        except Exception as err:  # pylint:disable=broad-exception-caught
            LOGGER.exception("Background task generated an exception")
            errors.append(err)

    if errors:
        sys.exit(1)

    LOGGER.info("Done")


if __name__ == '__main__':
    main()



""" Common custom widgets """

import logging
import os
import os.path

from PySide6.QtCore import QMargins, QPoint, QRect, QSize, Qt
from PySide6.QtGui import QColor
from PySide6.QtWidgets import (QColorDialog, QDialog, QFileDialog, QHBoxLayout,
                               QLabel, QLayout, QLineEdit, QPlainTextEdit,
                               QPushButton, QSizePolicy, QVBoxLayout, QWidget)

from .. import util
from .file_utils import FileRole

LOGGER = logging.getLogger(__name__)


def wrap_layout(layout: QLayout):
    """ Wrap a layout in a QWidget, which is for some reason not a standard Qt function """
    widget = QWidget()
    widget.setLayout(layout)
    return widget


class FileSelector(QWidget):
    """ A file selector textbox with ... button """

    def __init__(self, role: FileRole, album_editor=None, text=''):
        super().__init__()

        layout = QHBoxLayout()
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)

        self.setLayout(layout)

        self.role = role
        self.album_editor = album_editor
        self.file_path = QLineEdit(text)
        self.button = QPushButton("...")

        layout.addWidget(self.file_path)
        layout.addWidget(self.button)

        self.button.clicked.connect(self.choose_file)

    def choose_file(self):
        """ Pick a file """

        path = self.file_path.text()
        if self.album_editor:
            start_dir = self.album_editor.get_last_directory(self.role, path)
        elif self.file_path and os.path.isabs(path):
            start_dir = os.path.dirname(path)
        else:
            start_dir = self.role.default_directory

        LOGGER.debug("start_dir=%s filter=%s",
                     start_dir, self.role.file_filter)
        (filename, _) = QFileDialog.getOpenFileName(self,
                                                    f'Select your {self.role.name}',
                                                    start_dir,
                                                    self.role.file_filter)
        if filename:
            # Update the global default for files of this role
            filedir = os.path.dirname(filename)
            self.role.default_directory = filedir

            if self.album_editor:
                # Update the album editor's default for files of this role
                self.album_editor.set_last_directory(self.role, filedir)
                filename = util.make_relative_path(
                    self.album_editor.filename)(filename)
                os.path.dirname(filename)

            self.file_path.setText(filename)

    def text(self):
        """ Get the value out """
        return self.file_path.text()

    def setText(self, text):
        """ Set the value """
        # pylint:disable=invalid-name
        return self.file_path.setText(text)


class FlowLayout(QLayout):
    """ Layout with reflow
    adapted from https://doc.qt.io/qtforpython-6/examples/example_widgets_layouts_flowlayout.html
    """
    # pylint:disable=invalid-name,missing-function-docstring

    def __init__(self, parent=None):
        super().__init__(parent)

        if parent is not None:
            self.setContentsMargins(QMargins(0, 0, 0, 0))

        self._item_list = []

    def __del__(self):
        item = self.takeAt(0)
        while item:
            item = self.takeAt(0)

    def addItem(self, item):
        self._item_list.append(item)

    def count(self):
        return len(self._item_list)

    def itemAt(self, index):
        if 0 <= index < len(self._item_list):
            return self._item_list[index]

        return None

    def takeAt(self, index):
        if 0 <= index < len(self._item_list):
            return self._item_list.pop(index)

        return None

    def expandingDirections(self):
        return Qt.Orientation(0)

    def hasHeightForWidth(self):
        return True

    def heightForWidth(self, width):
        height = self._do_layout(QRect(0, 0, width, 0), True)
        return height

    def setGeometry(self, rect):
        super().setGeometry(rect)
        self._do_layout(rect, False)

    def sizeHint(self):
        return self.minimumSize()

    def minimumSize(self):
        size = QSize()

        for item in self._item_list:
            size = size.expandedTo(item.minimumSize())

        size += QSize(2 * self.contentsMargins().top(),
                      2 * self.contentsMargins().top())
        return size

    def _do_layout(self, rect, test_only):
        x = rect.x()
        y = rect.y()
        line_height = 0
        spacing = self.spacing()

        for item in self._item_list:
            style = item.widget().style()
            layout_spacing_x = style.layoutSpacing(
                QSizePolicy.ControlType.PushButton,
                QSizePolicy.ControlType.PushButton,
                Qt.Orientation.Horizontal
            )
            layout_spacing_y = style.layoutSpacing(
                QSizePolicy.ControlType.PushButton,
                QSizePolicy.ControlType.PushButton,
                Qt.Orientation.Vertical
            )
            space_x = spacing + layout_spacing_x
            space_y = spacing + layout_spacing_y
            next_x = x + item.sizeHint().width() + space_x
            if next_x - space_x > rect.right() and line_height > 0:
                x = rect.x()
                y = y + line_height + space_y
                next_x = x + item.sizeHint().width() + space_x
                line_height = 0

            if not test_only:
                item.setGeometry(QRect(QPoint(x, y), item.sizeHint()))

            x = next_x
            line_height = max(line_height, item.sizeHint().height())

        return y + line_height - rect.y()


class ColorPicker(QWidget):
    """ A color picker button """

    def __init__(self, parent=None, label=""):
        super().__init__(parent)

        self.label = label

        self._value = QColor.fromString('#000000')
        self._button = QPushButton()
        self._button.setAutoFillBackground(True)

        self.setName('#000000')

        self._button.clicked.connect(self.pickColor)

        hbox = QHBoxLayout()
        hbox.setContentsMargins(0, 0, 0, 0)
        hbox.setSpacing(0)
        hbox.addWidget(self._button)
        self.setLayout(hbox)

    def pickColor(self):  # pylint:disable=invalid-name
        """ Pick a color """
        color = QColorDialog.getColor(self._value)
        if color.isValid():
            self.setValue(color)

    def value(self):
        """ Get the color as a Qt color object """
        return self._value

    def name(self):
        """ Get the color as a hex string """
        return self._value.name()

    def setValue(self, color):  # pylint:disable=invalid-name
        """ Set the color value by Qt color """
        self._value = color
        fg_color = 'white' if self._value.valueF() < 0.5 else 'black'
        self._button.setText(
            f'{self.label}: {color.name()}' if self.label else color.name())
        self._button.setStyleSheet(
            f'background-color: {color.name()}; color: {fg_color};')

    def setName(self, name):  # pylint:disable=invalid-name
        """ Set the color value by hex string """
        self.setValue(QColor.fromString(name))


class ErrorMessage(QDialog):
    """ An error box with details; the builtin QMessageBox behaves strangely """

    def __init__(self, parent, errors):
        super().__init__(parent)
        self.setWindowTitle("Error")
        self.setWindowModality(Qt.WindowModality.WindowModal)

        layout = QVBoxLayout()

        if len(errors) == 1:
            layout.addWidget(QLabel("An error occurred:"))
        else:
            layout.addWidget(QLabel(f"{len(errors)} errors occurred:"))

        text = QPlainTextEdit()
        text.setPlainText('\n\n'.join(str(e) for e in errors))
        text.setReadOnly(True)
        layout.addWidget(text)

        button = QPushButton("Okay")
        button.clicked.connect(self.accept)
        layout.addWidget(button)

        self.setLayout(layout)



""" Player engine for Camptown """

import copy
import logging
import os.path
import shutil
import typing

import camptown

LOGGER = logging.getLogger(__name__)


class Player:
    """ A player using `Camptown <https://github.com/fluffy-critter/camptown>`_. """

    def __init__(self, art_size=200, fullsize_art_size=65536):
        self.art_size = art_size
        self.fullsize_art_size = fullsize_art_size

    @property
    def name(self):
        return 'camptown'

    @property
    def art_rendition_sizes(self):
        return (
            ("1x", self.art_size),
            ("2x", self.art_size*2),
            ("fullsize", self.fullsize_art_size)
        )

    def convert(self, input_dir, output_dir, album, protections, **vars):
        """ Convert a Bandcrash album spec to a Camptown player """
        info: dict[str, typing.Any] = {'tracks': []}

        for key, tgt in (
            ('title', 'title'),
            ('artist', 'artist'),
            ('artwork_preview', 'artwork'),
            ('artist_url', 'artist_url'),
            ('album_url', 'album_url'),
        ):
            if key in album:
                info[tgt] = album[key]

        for track in album['tracks']:
            out: dict[str, typing.Any] = {}
            for key, tgt in (
                ('preview_mp3', 'filename'),
                ('artist', 'artist'),
                ('title', 'title'),
                ('explicit', 'explicit'),
                ('artwork_preview', 'artwork'),
                ('lyrics', 'lyrics'),
                ('about', 'about'),
                ('duration', 'duration'),
            ):
                if key in track:
                    out[tgt] = track[key]
            if track.get('group'):
                out['title'] = f"{track['group']}: {track.get('title', '(no title)')}"
            info['tracks'].append(out)

        theme = album.get('theme', {})
        info['theme'] = copy.deepcopy(theme)

        if 'user_css' in theme:
            shutil.copy(os.path.join(input_dir, info['theme']['user_css']),
                        os.path.join(output_dir, 'user.css'))
            info['theme']['user_css'] = 'user.css'
            protections.add('user.css')

        files = camptown.process(info, output_dir,
                                 footer_urls=[
                                     ('https://fluffy.itch.io/bandcrash', 'Bandcrash')],
                                 file_callback=lambda path: os.path.join(input_dir, path))

        protections |= files
